import java.util.*;
import java.io.*;
import java.net.*;
import java.text.DateFormat;

// jena
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.util.iterator.*;
import com.hp.hpl.jena.vocabulary.RDF;

// misc
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.SuffixFileFilter;


/**
 * Converter from arbitrary RDF to file system (with symbolic links).
 * 
 * @author Michael Sintek
 */
public class RDF2FS {

  Model model;
  PrintWriter writer;
  HashMap<String, String> namespaces; // namespace -> abbrev
  HashSet<String> abbrevs;
  HashMap<Resource, String> fileNames; // Resource -> fileName (usually qnames)
  int anonCounter = 0;

  static final String ENCODING = "ISO-8859-1"; // for executable script

  public void toFS(Model model, PrintWriter writer) throws Exception {
    this.model = model;
    this.writer = writer;
    toFS();
  }

  void toFS() throws Exception {
    namespaces = new HashMap();
    abbrevs = new HashSet();
    fileNames = new HashMap();
    // walk over all statements
    StmtIterator stmtIter = model.listStatements();
    while (stmtIter.hasNext()) {
      Statement stmt = stmtIter.nextStatement();
      Resource subject = stmt.getSubject();
      Property predicate = stmt.getPredicate();
      RDFNode object = stmt.getObject();
      String subjectFilename = getFileName(subject, true);
      String predicateFilename = getFileName(predicate, false);
      if (object.isResource()) {
        String objectFilename = getFileName((Resource)object, true);
        // add symbolic link
        String predicateDirectory = subjectFilename + "/" + predicateFilename;
        writer.println("mkdir " + predicateDirectory + " > /dev/null 2>&1");
        writer.println("ln -s ../../" + objectFilename + " " + predicateDirectory);
      } else { // literal
        String predicateLiteralFile = subjectFilename + "/" + predicateFilename;
        String objectString = ((Literal)object).getString();
        objectString = objectString.replace("'", "'\"'\"'"); // ' -> '"'"'
        writer.println("echo '" + objectString + 
          "' >> " + predicateLiteralFile);
      }
    }
  }

  String getFileName(Resource resource, boolean create) throws Exception {
    String name = fileNames.get(resource);
    if (name != null)
      return name;
    else {
      name = getFileName(resource);
      fileNames.put(resource, name);
      if (create)
        writer.println("mkdir " + name);
      return name;
    }
  }

  String getFileName(Resource resource) throws Exception {
    // returns nsabbrev#ln (we use # and not : to create usable filenames)
    // (and sometimes simple file names)
    String uri = resource.getURI();
    if (uri == null) { // anon
      String id = resource.getId().toString();
      if (id != null && id != "") {
        if (id.startsWith("-") || Character.isDigit(id.charAt(0)))
          id = "anon" + id; // ???
        return getFilenameComponent(id);
      } else
        return getAnonymousName();
    } else { // create nice qname
      // split uri at last # (or /, or :, if there is no #)
      int i = uri.lastIndexOf('#');
      if (i == -1) i = uri.lastIndexOf('/');
      if (i == -1) i = uri.lastIndexOf(':');
      if (i >= 0) { // we have found a namespace
        String namespace = uri.substring(0, i);
        String localname = uri.substring(i+1);
        // now let's find a namespace abbreviation
        String nsabbrev = null;
        nsabbrev = namespaces.get(namespace);
        if (nsabbrev == null) {
          // first, we check for common namespaces (TODO: make configurable)
          if (namespace.indexOf("rdf-syntax-ns") >= 0) nsabbrev = "rdf";
          else if (namespace.indexOf("rdf-schema") >= 0) nsabbrev = "rdfs";
          // else if (ns.indexOf("XMLSchema") >= 0) ns = "xsd"; ??
          else if (namespace.indexOf("purl.org/dc/elements") >= 0) nsabbrev = "dc";
          else if (namespace.indexOf("purl.org/rss") >= 0) nsabbrev = "rss";
          else if (namespace.indexOf("xmlns.com/foaf") >= 0) nsabbrev = "foaf";
          else if (namespace.indexOf("xmlns.com/wot") >= 0) nsabbrev = "wot";
          else if (namespace.indexOf("xmlns.com/wordnet") >= 0) nsabbrev = "wordnet";
          else if (namespace.indexOf("inkel/rdf/schemas/lang") >= 0) nsabbrev = "lang";
          if (nsabbrev == null) {
            // take last component of namespace
            i = namespace.lastIndexOf('/');
            if (i >= 0) nsabbrev = namespace.substring(i+1);
            else nsabbrev = namespace;
          }
          // do some corrections of nsabbrev
          if (nsabbrev.equals("") || nsabbrev.startsWith("-") 
                                  || Character.isDigit(nsabbrev.charAt(0)))
            nsabbrev = "ns" + nsabbrev; // ???
          // check for duplicates (and rename if necessary)
          while (abbrevs.contains(nsabbrev))
            nsabbrev += "_"; // TODO: append number
          // store abbrev and namespace
          namespaces.put(namespace, nsabbrev);
          abbrevs.add(nsabbrev);
        }
        // System.out.println(uri + " -> " + nsabbrev + " : " + localname);
        // finally, let's create the file name
        //if (localname.equals("") || localname.startsWith("-") 
        //                         || Character.isDigit(localname.charAt(0)))
        //  localname = "res" + localname; // ???
        String name = getFilenameComponent(nsabbrev) + "#" + 
                      getFilenameComponent(localname);
        return name;
      } else { // no qname (should never happen, of course)
        String name = uri;
        if (name.startsWith("-") || Character.isDigit(name.charAt(0)))
          name = "res" + name; // ???
        return getFilenameComponent(name);
      }
    }
  }

  String getFilenameComponent(String string) throws Exception {
    return URLEncoder.encode(string, ENCODING);
  }

  String getAnonymousName() {
    anonCounter++;
    return "_" + anonCounter; // check for duplicates!!
  }


  // static methods

  public static void toFS(File file) throws Exception {
    String lang = "RDF/XML";
    if (file.getPath().endsWith(".n3")) lang = "N3";
    toFS(file, lang);
  }

  public static void toFS(File file, String lang) throws Exception {
    String fileName = file.getPath();
    System.out.println("handling " + fileName);
    String shFileName = fileName + ".sh";
    Model model = ModelFactory.createDefaultModel();
    FileInputStream is = new FileInputStream(file);
    model.read(is, "http://www.example.com/" + fileName + "#" , lang);
    is.close();
    PrintWriter writer =
      new PrintWriter(new BufferedWriter(new OutputStreamWriter(
        new FileOutputStream(shFileName), ENCODING)));
    writer.println("# RDF2FS for " + file.getName());
    writer.println("# generated on " + 
        DateFormat.getDateTimeInstance().format(new Date()));
    writer.println();
    (new RDF2FS()).toFS(model, writer);
    writer.close();
    System.out.println("created " + shFileName);
  }
  
  public static void main(String[] args) {
    boolean stackTrace = false;
    try {
      for (int i = 0; i < args.length; i++) {
        String fileName = args[i];
        if (fileName.equals("-st")) {
          stackTrace = true;
          continue;
        }
        File file = new File(fileName);
        if (file.isDirectory()) {
          FileFilter fileFilter = new SuffixFileFilter(".rdf");
          File[] files = file.listFiles(fileFilter);
          for (int j = 0; j < files.length; j++) { toFS(files[j], "RDF/XML"); }
          fileFilter = new SuffixFileFilter(".n3");
          files = file.listFiles(fileFilter);
          for (int j = 0; j < files.length; j++) { toFS(files[j], "N3"); }
        } else {
          toFS(file);
        }
      }
    } catch (Exception e) {
      if (stackTrace) 
        e.printStackTrace(System.out);
      else
        System.out.println(e);
    }
  }

}

