#!/usr/bin/env python 

# python standard libs
import sys
import os
import os.path
import re
import getopt
import time
import urllib

# From - http://rdflib.net, we need version 2.4
import rdflib

DEBUG=False

def formatFromFile(f):
    if f.endswith(".n3"):
        return "n3"

    if f.endswith(".nt"):
        return "nt"

    if f.endswith("trix"):
        return "trix"

    return "xml" 


ns={}
ns["http://xmlns.com/foaf/0.1/"]="foaf"
ns["http://purl.org/dc/elements/1.1/"]="dc"
ns["http://www.w3.org/1999/02/22-rdf-syntax-ns#"]="rdf"
ns["http://www.w3.org/2000/01/rdf-schema#"]="rdfs"

# Keep track of new NS mappings we create
otherns=0

# base URI of current doc
base=None

def splitURI(uri):
    """Return namespace/localname tuple"""
    m=re.search("(.*(?:/|#))([^/]+?)?$",uri)
    if m:
        #if DEBUG:
            #sys.stderr.write("%s => (%s, %s)\n"%(uri,m.groups()[0],m.groups()[1]))
        if m.groups()[0] in ["http://", "https://", "ftp://" ]:
            return ()
            
        return m.groups()
    else:
        return ()

def toLiteral(lit):
    return lit.replace("'", "\'\"\'\"\'").replace("\n","\\n")

def findAbbrev(s):
    global otherns

    candidates=filter(lambda x: len(x)>0, re.split("[^a-z]",s.lower()))

    # remove the protocols and common tld. 
    candidates=filter(lambda x: x not in ["http","ftp","mailto","urn","com","net","org"], candidates)

    candidates.reverse()

    oldkeys=set(ns.values())

    if len(candidates)>0:
        for c in candidates:
            if c not in oldkeys:
                return c

        base=candidates[0]

    else:
        # no candidates found, go with "ns"
        sys.stderr.write("No candidate NS abbrev found for %s\n"%s)
        base="ns"
        
    i=0
    c="%s%i"%(base,i)
    while c in oldkeys:
        i+=1
        c="%s%i"%(base,i)

    return c

def toFileName(resource):    
    """Returns the string that is the filename of this resource on disk"""

    global ns

    if isinstance(resource, rdflib.BNode):
        return str(resource)

    if resource.startswith("mailto:"):
        return str(resource)

    s=splitURI(resource)
    if s==():
        if resource.startswith("http") or \
           resource.startswith("ftp") or \
           resource.startswith("https"):
            return str(resource)
        else: 
            sys.stderr.write("Warning, no NS found for %s and no protocol, assuming relative URI\n"%resource)
            return "%s#%s"%(base,resource)

    if s[0]=="xml#":
        # rdflib weirdness
        return "%s#%s"%(base,s[1])
    
    if s[0] not in ns:
        # new NS, create a mapping now
        abbrev=findAbbrev(s[0])
        if DEBUG: sys.stderr.write("Using %s as prefix for %s\n"%(abbrev, s[0]))
        ns[s[0]]=abbrev

        
    return "%s#%s"%(ns[s[0]],s[1])
                
def toFile(resource):
    return urllib.quote(toFileName(resource)).replace("%23","#").replace("/","%2F")
    


def toFS(f):

    """Given a filename for a rdf file, create filename.sh with a script for
    creating the RDF links/folders/etc."""
    
    global base

    base=f
    created=set()
    
    g=rdflib.Graph()
    sys.stderr.write("Loading %s..."%f)
    try:
        
        g.load(f, formatFromFile(f))
    except e:
        sys.stderr.write("[FAIL] Could not load file %s: %s"%(f,e))

        return 
    sys.stderr.write("[DONE] Loaded %i triples.\n"%len(g))
    
    script=file(f+".sh", "w")
    script.write("# RDF2FS, from %s\n"%f)
    script.write("# Generated on : %s\n"%time.asctime())
    
    for s,p,o in g:
        sfile=toFile(s)
        pfile=toFile(p)

        if sfile not in created:
            script.write("mkdir %s\n"%sfile)
            created.add(sfile)            
        
        if isinstance(o,rdflib.Literal):
            # literal
            pfile="%s/%s"%(sfile,toFile(p))            
            script.write("echo '%s'>> %s\n"%(toLiteral(o),pfile))
        else:            
            # resource
            ofile=toFile(o)
            if ofile not in created:
                script.write("mkdir %s\n"%ofile)
                created.add(ofile)

            pfile="%s/%s"%(sfile,toFile(p))
            
            if pfile not in created: 
                script.write("mkdir %s\n"%pfile)
                created.add(pfile)

            script.write("ln -s ../../%s %s/%s\n"%(ofile,pfile,ofile))


    for namespace,prefix in ns.iteritems():
        script.write("echo '%s=>%s'>> .namespace_mappings\n"%(prefix,namespace))
        
    script.close()

    
    

if __name__=='__main__': 
    args,files=getopt.getopt(sys.argv[1:],"d")

    args=dict(args)
    if args.has_key("-d"):
        DEBUG=True
    
    for f in files:
        if os.path.isdir(f):
            for f2 in os.listdir(f):
                toFS(f2)
        else:
            toFS(f)
